#include "dhcp.h"
#include "dhcpcd.h"


void DumpPacket(char *pPacket, int paketLen);

// Don't forget the following next two lines before using this function
//		WSADATA			Data;
//		int status = WSAStartup(MAKEWORD(1,1), &Data);
// This function is a reverse function of inet_addr
bool
rev_inet_addr(__u32 netByteIPAddr32, char *dotIP)
{
	memset(dotIP, 0, 16);
	
	struct in_addr	addr;
	addr.s_addr = netByteIPAddr32;

	memcpy(dotIP, inet_ntoa(addr), 16); 
	return true;
}


bool getAdapterInfo2(
	const char* nicName,		// input
	char* AdapterName,			// output
	unsigned char*	macAddr,	// output	
	char* ipAddress,			// output	dot-IP			
	char* maskAddress			// output	dod-IP		
)
{
	// string that contains a list of the network adapters
	WCHAR		TempBuffer[1024]; 

	char        adapterList[512];
	char        adapterList2[512];

	char		AdapterName2[512];

	__u32		ipAddr;				// IP Address
	__u32		maskAddr;			// Mask (Subnet) Address

	DWORD dwVersion=GetVersion();
	DWORD dwWindowsMajorVersion =  (DWORD)(LOBYTE(LOWORD(dwVersion)));

	// Windows NT 
	if (!(!(dwVersion >= 0x80000000 && dwWindowsMajorVersion >= 4)))
		return false;
  
	ULONG		AdapterLength = 1024;

	memset(TempBuffer,0, sizeof(TempBuffer));
	memset(adapterList,0, sizeof(adapterList));
	memset(adapterList2,0, sizeof(adapterList2));

	WCHAR		*temp,*temp1;

	int nicNameLen = strlen(nicName);
	temp = (WCHAR *)(adapterList);
	
	for(int k=0; k < nicNameLen; k++)
	{
		*temp++ = *nicName++;
	}

	// Change adapterList the part of "/" to "//" for Windows NT

	
	memcpy(adapterList2,adapterList,sizeof(adapterList));
	temp1 = temp = (WCHAR *)(adapterList2);
	
	int adapterList2Length = 0;
	while ((*temp!='\0')||(*(temp-1)!='\0'))
	{
		adapterList2Length++;
		temp++;
	}
	
	int j =0;
	temp = temp1;
	while ((*temp!='\0')||(*(temp-1)!='\0'))
	{
		if (*temp == 0x005c)
		{
			int k;
			for (k = adapterList2Length ; k > j ; k--)
			{
				*(temp1 + k + 1) = *(temp1 + k); 
			}
			
			*(temp + 1) =  0x005c;	// add "/"
			temp++;
			adapterList2Length++;
		}
		j++;
		temp++;
	}

	
	// Get the IP Address
	BOOLEAN truefalse = 
		PacketGetNetInfo(adapterList2,(PULONG)(&ipAddr),(PULONG)(&maskAddr));
	
	rev_inet_addr(htonl(ipAddr), ipAddress);
	rev_inet_addr(htonl(maskAddr), maskAddress);


	memcpy(AdapterName, adapterList, 512 /*sizeof(AdapterName) */);
	memcpy(AdapterName2, adapterList2, 512 /*sizeof(AdapterName2) */);
	
	LPADAPTER temp_lpAdapter 
		= PacketOpenAdapter(AdapterName);
	
	if (!temp_lpAdapter || (temp_lpAdapter->hFile == INVALID_HANDLE_VALUE))
	{
		DWORD dwErrorCode=GetLastError();
		
		return false;
	}
	
	// Check the Adapter Type (It should be Ethernet)
	NetType netType;
	
	PacketGetNetType(temp_lpAdapter, &netType);
	if (netType.LinkType != NdisMedium802_3)
		return false;
	
	// Get the MAC Address
	PACKET_OID_DATA pod;
	
	pod.Oid = OID_802_3_PERMANENT_ADDRESS;
	pod.Length = ETH_ALEN /* sizeof(macAddr) */;
	
	if (PacketRequest(temp_lpAdapter, 0, &pod) == NULL)
	{
		return false;
	}
	
	memcpy(macAddr, pod.Data, pod.Length);
	
	PacketCloseAdapter(temp_lpAdapter);
	
	return true;
}

unsigned short 
in_cksum(unsigned short *addr, int len)
{
	int				nleft = len;
	int				sum = 0;
	unsigned short	*w = addr;
	unsigned short	answer = 0;

	while (nleft > 1)
	{
		sum += *w++;
		nleft -= 2;
	}

	if (nleft == 1)
	{
		*(unsigned char *)(&answer) = *(unsigned char *)w;
		sum += answer;
	}

	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

	answer = ~sum;

	return(answer);
};

unsigned short
udpchecksum(
	struct udphdr *_udphdr, 
	struct iphdr *_iphdr, 
	struct dhcpMessage *_dhcpMsg,
	int udpDataLenInPacket)
{
	_udphdr->check = 0;  // Checksum should be set "0" first.

	unsigned char buf[MAX_ETH_SIZE];
	memset(buf, 0, MAX_ETH_SIZE);
	
	int buflen = 0;

	struct pesudeIphdr _pesudeIphdr;

	// make pesude IP header
	_pesudeIphdr.saddr = _iphdr->saddr;
	_pesudeIphdr.daddr = _iphdr->daddr;
	_pesudeIphdr.zero = 0;	// should be 0
	_pesudeIphdr.protocol = _iphdr->protocol;
	_pesudeIphdr.len = _udphdr->len;

	memcpy(buf, &_pesudeIphdr, sizeof(struct pesudeIphdr));
	
	buflen += sizeof(struct pesudeIphdr);

	// copy packet for udp to buf
	memcpy(buf + buflen, _udphdr, SIZE_OF_UDPHDR);
	buflen += SIZE_OF_UDPHDR;

	unsigned char *testP = (unsigned char *)_dhcpMsg;

	memcpy((buf+buflen), testP , udpDataLenInPacket);
	buflen += udpDataLenInPacket;

	return in_cksum((unsigned short *)buf, buflen);
}

unsigned int __stdcall _dhcpcdThread(void* t) 
{
	CDhcpcd* dhcpcd = (CDhcpcd*)t;
	dhcpcd->dhcpClientDemon();
	printf("_dhcpcdThread is destructed..\n");

	_endthreadex(0);
	return 0;
}


bool
CDhcpcd::buildDiscoverMsg(char *_packetMessage)
{
	////////// Making DHCP (DISCOVER) message
		
	struct ethhdr	_ethhdr;
	struct iphdr	_iphdr;
	struct udphdr	_udphdr;
	struct dhcpMessage _dscpMsg;


	///// Ethernet Header

	// Destination: ff:ff:ff:ff:ff:ff (ff:ff:ff:ff:ff:ff)
	_ethhdr.h_dest[0]=0xff;
	_ethhdr.h_dest[1]=0xff;
	_ethhdr.h_dest[2]=0xff;
	_ethhdr.h_dest[3]=0xff;
	_ethhdr.h_dest[4]=0xff;
	_ethhdr.h_dest[5]=0xff;	

	// Source:
	memcpy(_ethhdr.h_source, m_macAddr, sizeof(m_macAddr));

	// Type
	_ethhdr.h_proto = htons(ETH_P_IP);

	///// IP Header
	_iphdr.ihl = 5;				// Header length 20 byte
	_iphdr.version = 4;
	_iphdr.tos = 0;
	_iphdr.tot_len = htons(328);  // Fixed
	_iphdr.id = htons(0x0000);
	_iphdr.frag_off = htons(0x00);
	_iphdr.ttl = 128;		// Fixed
	_iphdr.protocol = UDP;
	_iphdr.saddr = htonl(0);
	_iphdr.daddr = htonl(0xffffffff);
	// IP checksum
	_iphdr.check = 0;
	_iphdr.check = in_cksum((unsigned short *)(&_iphdr), SIZE_OF_IPHDR);

	///// UDP Header
	_udphdr.source = htons(DHCP_CLIENT_PORT); // Fixed
	_udphdr.dest = htons(DHCP_SERVER_PORT);	// Fixed
	_udphdr.len = htons(308);	// Fixed

	////// Bootstrap
	_dscpMsg.op = DHCP_BOOTREQUEST;
	_dscpMsg.htype = ARPHRD_ETHER;
	_dscpMsg.hlen = ETH_ALEN;		/* hardware address length */
	_dscpMsg.hops = 0;	/* should be zero in client's message */

	srand( (unsigned)time( NULL ) );

	m_xid = (rand() + 1) * (rand() + 1);
	_dscpMsg.xid = htonl(m_xid);
	_dscpMsg.secs = htons(0);
	_dscpMsg.flags	=	htons(0);

	_dscpMsg.ciaddr = htons(0);
	_dscpMsg.yiaddr = htons(0);
	_dscpMsg.siaddr = htons(0);		/* should be zero in client's messages */
	_dscpMsg.giaddr = htons(0);		/* should be zero in client's messages */
	
	memset(_dscpMsg.chaddr, 0, sizeof(_dscpMsg.chaddr));
	memcpy(_dscpMsg.chaddr,m_macAddr,ETH_ALEN);
	memset(_dscpMsg.sname, 0, sizeof(_dscpMsg.sname));
	memset(_dscpMsg.file, 0, sizeof(_dscpMsg.file));

	memset(_dscpMsg.options, 0, sizeof(_dscpMsg.options));  
	
	__u32 magic_cookie = htonl(MAGIC_COOKIE); 
	memcpy(_dscpMsg.options,&magic_cookie,4);

	register unsigned char *p = _dscpMsg.options + 4;

	*p++ = dhcpMessageType;		// 53	&35
	*p++ = 1;					// 1	&01
	*p++ = DHCP_DISCOVER;		// 1	&01

	int dummy = strlen(m_dhcpID);

	*p++ = dhcpClientIdentifier;// 61	&3D
	*p++ = ETH_ALEN + 1 + strlen(m_dhcpID) + 1; 
	*p++ = ARPHRD_ETHER;		// type: Ethernet address
	memcpy(p,m_macAddr,ETH_ALEN);							
	p += ETH_ALEN;
	memcpy(p,":",1);
	p += 1;
	memcpy(p, m_dhcpID, strlen(m_dhcpID));
	p += strlen(m_dhcpID);


	char HostName[255];
	gethostname(HostName, 255);
	*p++ = hostName;			// 12	&0C

	int HostName_len;
	*p++ = HostName_len = strlen(HostName) + 1;	 	
	memset(p,0, HostName_len);
	memcpy(p,HostName,HostName_len -1);
	p += HostName_len;
	
	*p++ = dhcpParamRequest;	// 55	&37
	*p++ = 4;					// Four Parameters	
	*p++ = subnetMask;			// 1	&01
	*p++ = routersOnSubnet;		// 3	&03
	*p++ = dns;					// 6	&06
	*p++ = domainName;			// 15	&0F
	
	*p = endOption;				// 255	&FF

	// UDP checksum
	int udpDataLenInPacket = 
		ntohs(_iphdr.tot_len) - SIZE_OF_IPHDR - SIZE_OF_UDPHDR;
	_udphdr.check = 0;  // Checksum should be set "0" first.
	_udphdr.check = 
			udpchecksum(&_udphdr, &_iphdr, &_dscpMsg, udpDataLenInPacket);	

	///// Send DISCOVER Message
	char *_pPacket = _packetMessage;

	memcpy(_pPacket, &_ethhdr, sizeof(_ethhdr));
	_pPacket += sizeof(_ethhdr);

	memcpy(_pPacket, &_iphdr, sizeof(_iphdr));
	_pPacket += sizeof(_iphdr);

	memcpy(_pPacket, &_udphdr, sizeof(_udphdr));
	_pPacket += sizeof(_udphdr);

	memcpy(_pPacket, &_dscpMsg, sizeof(_dscpMsg));

	return true;
}


bool
CDhcpcd::buildRequestMsg(
	char *_packetMessage,
	struct dhcpGetInfo *_dhcpGetInfo
)
{
	////////// Making DHCP (DISCOVER) message
		
	struct ethhdr	_ethhdr;
	struct iphdr	_iphdr;
	struct udphdr	_udphdr;
	struct dhcpMessage _dscpMsg;


	///// Ethernet Header

	// Destination: ff:ff:ff:ff:ff:ff (ff:ff:ff:ff:ff:ff)
	_ethhdr.h_dest[0]=0xff;
	_ethhdr.h_dest[1]=0xff;
	_ethhdr.h_dest[2]=0xff;
	_ethhdr.h_dest[3]=0xff;
	_ethhdr.h_dest[4]=0xff;
	_ethhdr.h_dest[5]=0xff;	

	// Source:
	memcpy(_ethhdr.h_source, m_macAddr, sizeof(m_macAddr));

	// Type
	_ethhdr.h_proto = htons(ETH_P_IP);

	///// IP Header
	_iphdr.ihl = 5;				// Header length 20 byte
	_iphdr.version = 4;
	_iphdr.tos = 0;
	_iphdr.tot_len = htons(328);  // Fixed
	_iphdr.id = htons(0x0000);
	_iphdr.frag_off = htons(0x00);
	_iphdr.ttl = 128;		// Fixed
	_iphdr.protocol = UDP;
	_iphdr.saddr = htonl(0);
	_iphdr.daddr = htonl(0xffffffff);
	// IP checksum
	_iphdr.check = 0;
	_iphdr.check = in_cksum((unsigned short *)(&_iphdr), SIZE_OF_IPHDR);

	///// UDP Header
	_udphdr.source = htons(DHCP_CLIENT_PORT); // Fixed
	_udphdr.dest = htons(DHCP_SERVER_PORT);	// Fixed
	_udphdr.len = htons(308);	// Fixed


	////// Bootstrap
	_dscpMsg.op = DHCP_BOOTREQUEST;
	_dscpMsg.htype = ARPHRD_ETHER;
	_dscpMsg.hlen = ETH_ALEN;		/* hardware address length */
	_dscpMsg.hops = 0;	/* should be zero in client's message */

	srand( (unsigned)time( NULL ) );
	
	m_xid = (rand() + 1) * (rand() + 1);
	_dscpMsg.xid = htonl(m_xid);
	_dscpMsg.secs = htons(0);
	_dscpMsg.flags	=	htons(0);

	_dscpMsg.ciaddr = htons(0);
	_dscpMsg.yiaddr = htons(0);
	_dscpMsg.siaddr = htons(0);		/* should be zero in client's messages */
	_dscpMsg.giaddr = htons(0);		/* should be zero in client's messages */
	
	memset(_dscpMsg.chaddr, 0, sizeof(_dscpMsg.chaddr));
	memcpy(_dscpMsg.chaddr,m_macAddr,ETH_ALEN);
	memset(_dscpMsg.sname, 0, sizeof(_dscpMsg.sname));
	memset(_dscpMsg.file, 0, sizeof(_dscpMsg.file));

	memset(_dscpMsg.options, 0, sizeof(_dscpMsg.options));  
	
	__u32 magic_cookie = htonl(MAGIC_COOKIE); 
	memcpy(_dscpMsg.options,&magic_cookie,4);

	register unsigned char *p = _dscpMsg.options + 4;

	*p++ = dhcpMessageType;		// 53	&35
	*p++ = 1;					// 1	&01
	*p++ = DHCP_REQUEST;		// 3	&03

	*p++ = dhcpClientIdentifier;// 61	&3D
	*p++ = ETH_ALEN + 1 + strlen(m_dhcpID) + 1; 
	*p++ = ARPHRD_ETHER;		// type: Ethernet address
	memcpy(p,m_macAddr,ETH_ALEN);							
	p += ETH_ALEN;
	memcpy(p,":",1);
	p += 1;
	memcpy(p, m_dhcpID, strlen(m_dhcpID));
	p += strlen(m_dhcpID);

	char HostName[255];
	gethostname(HostName, 255);
	*p++ = hostName;			// 12	&0C

	int HostName_len;
	*p++ = HostName_len = strlen(HostName) + 1;	 	
	memset(p,0, HostName_len);
	memcpy(p,HostName,HostName_len -1);
	p += HostName_len;
	
	*p++ = dhcpRequestedIPaddr;	// 50	&32
	*p++ = 4;
	memcpy(p, &(_dhcpGetInfo->yourClientIPAddress), 4);
	p += 4;

	*p++ = dhcpServerIdentifier;	// 54	&36
	*p++ = 4;
	memcpy(p, &(_dhcpGetInfo->dhcpServerIdentifier), 4);
	p += 4;

	*p++ = dhcpParamRequest;	// 55	&37
	*p++ = 4;					// Four Parameters	
	*p++ = subnetMask;			// 1	&01
	*p++ = routersOnSubnet;		// 3	&03
	*p++ = dns;					// 6	&06
	*p++ = domainName;			// 15	&0F
	
	*p = endOption;				// 255	&FF

	// UDP checksum
	int udpDataLenInPacket = 
		ntohs(_iphdr.tot_len) - SIZE_OF_IPHDR - SIZE_OF_UDPHDR;
	_udphdr.check = 0;  // Checksum should be set "0" first.
	_udphdr.check = 
			udpchecksum(&_udphdr, &_iphdr, &_dscpMsg, udpDataLenInPacket);	

	///// Send DISCOVER Message
	char *_pPacket = _packetMessage;

	memcpy(_pPacket, &_ethhdr, sizeof(_ethhdr));
	_pPacket += sizeof(_ethhdr);

	memcpy(_pPacket, &_iphdr, sizeof(_iphdr));
	_pPacket += sizeof(_iphdr);

	memcpy(_pPacket, &_udphdr, sizeof(_udphdr));
	_pPacket += sizeof(_udphdr);

	memcpy(_pPacket, &_dscpMsg, sizeof(_dscpMsg));

	return true;
}




bool 
CDhcpcd::parseDhcpMsgRecv(
	struct dhcpMessage *_dhcpMsg,				// [in]
	struct dhcpGetInfo *_dhcpGetInfo			// [out]
)
{	
	_dhcpGetInfo->xid = _dhcpMsg->xid; // Transaction ID
	_dhcpGetInfo->yourClientIPAddress 
		= _dhcpMsg->yiaddr; /* 'your' client IP address */
	_dhcpGetInfo->nextServerIPAddress 
		= _dhcpMsg->siaddr;	/* should be zero in client's messages */	
	
	dhcpOptions DhcpOptions[10];	// Now I don't think the case 
									// that has more than ten parameters

	register u_char *p = _dhcpMsg->options+4;
	unsigned char *end = _dhcpMsg->options + sizeof(_dhcpMsg->options);

	int i = 0;
	bool flag = false;

	while( p < end )
	{
		switch ( *p )
		{
		case endOption: 
			flag = true;
			break;

		case padOption: 
			p++; break;

		default:
			if ( p[1] )
			{
				DhcpOptions[i].num = p[0];
				DhcpOptions[i].len = p[1];
				memcpy(DhcpOptions[i].val, &(p[2]), p[1]);

				p += 2 + p[1] ;
			}
		} // switch

		if ((++i >= 10) || (flag == true))
			break;
	}
	
	
	for (int k=0;k < i; k++)
	{
		switch ( DhcpOptions[k].num )
		{
		case 1: /* subnet mask */
			memcpy(_dhcpGetInfo->subnetMask, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
			
		case 3: /* routers on subnet */
			memcpy(_dhcpGetInfo->defaultGateway, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
			
		case 6: /* dns servers */
			memcpy(_dhcpGetInfo->dns, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
			
		case 15: /* Domain Name */
			memcpy(_dhcpGetInfo->domainName, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
			
		case 51:/* dhcpAddrLeaseTime */
			memcpy(_dhcpGetInfo->leaseTime, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
			
		case 53:/* dhcpMessageType */
			memcpy(_dhcpGetInfo->dhcpMessageType, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
			
		case 54:/* dhcpServerIdentifier */
			memcpy(_dhcpGetInfo->dhcpServerIdentifier, DhcpOptions[k].val,
				DhcpOptions[k].len);
			break;
		}
	}

	return true;
}

bool 
CDhcpcd::packetDelimiter(struct dhcpGetInfo *_dhcpGetInfo)
{
	ULONG	ulBytesReceived;
	char	*pChar;
	char	*buf;
	u_int	off=0;
	u_int	dataLen;
	struct	bpf_hdr *hdr;
	
	ulBytesReceived = m_recvLpPacket->ulBytesReceived;
	buf = (char *)m_recvLpPacket->Buffer;	
	
	off=0;

	while(off<ulBytesReceived){
		
		hdr=(struct bpf_hdr *)(buf+off);

		off += hdr->bh_hdrlen;
			
		pChar = (char*)(buf+off);  // Top Address of thr Raw Packet

		dataLen =hdr->bh_datalen;
		off = Packet_WORDALIGN(off+dataLen); // For padding

		analizeRawPacket(pChar, dataLen, _dhcpGetInfo);
	}

	return true;
}
		

bool 
CDhcpcd::analizeRawPacket(
	char* pChar, 
	int dataLen,
	struct dhcpGetInfo *_dhcpGetInfo)
{
//	DumpPacket(pChar, dataLen);
	
	char* _pChar = pChar;

	// Step.1 
	// If the packet is not IP packet, stop it.
	
	struct ethhdr *_ethhdr;
	_ethhdr = (struct ethhdr *)_pChar;

	if (_ethhdr->h_proto != htons(ETH_P_IP))
	{
		return false;
	}

	// Step.3
	// Pick out each element of IP Packet

	struct iphdr *_iphdr;
	_pChar += SIZE_OF_ETHHDR; 	

	_iphdr = (struct iphdr *)_pChar;

	if (_iphdr->protocol != UDP)
	{
		return false;
	}

	// Step.4
	// Pick up UDP header
	
	struct udphdr *_udphdr;
	_pChar += SIZE_OF_IPHDR; 
	
	_udphdr = (struct udphdr *)_pChar;

	if ((ntohs(_udphdr->source) != DHCP_SERVER_PORT) || 
		(ntohs(_udphdr->dest) != DHCP_CLIENT_PORT))
	{
		return false;
	}

	_pChar += SIZE_OF_UDPHDR;
	parseDhcpMsgRecv((dhcpMessage *)_pChar, _dhcpGetInfo);

	////// Using by this _dhcpGetInfo, check if this dhcp packet is available.
	/// Filtering as follow

	if ((ntohl(_dhcpGetInfo->xid) == m_xid) 
		&& (_dhcpGetInfo->dhcpMessageType[0] == m_waitingPacketType))
	{
		memcpy(&m_dhcpGetInfo, _dhcpGetInfo, sizeof(m_dhcpGetInfo));	
		SetEvent(m_dhcpcdLock_out);

		return true;
	}
	return true;	
}



CDhcpcd::CDhcpcd(
//	char	*myAddress,
	char	*myNicName,
	char	*dhcpID)
{
	WSADATA			Data;
	int status = WSAStartup(MAKEWORD(1,1), &Data);

	m_dhcpcdLock_out = CreateEvent(0, FALSE, FALSE, 0);
	m_dhcpcdLock_in = CreateEvent(0, FALSE, FALSE, 0);

	m_CDhcpFlag = false;

	//////////////////// WinPcap Initialize

	/////////////// Adapter to send messages 

	if (getAdapterInfo2(myNicName, m_AdapterName, m_macAddr, m_ipAddr, m_maskAddr))
	{
		m_sendLpAdapter = PacketOpenAdapter(m_AdapterName);
		m_recvLpAdapter = PacketOpenAdapter(m_AdapterName);
	}
	else
		return;

	// PacketSetHwFilter(m_sendLpAdapter,NDIS_PACKET_TYPE_DIRECTED);
	PacketSetHwFilter(m_sendLpAdapter,NDIS_PACKET_TYPE_PROMISCUOUS);

	if((m_sendLpPacket = PacketAllocatePacket())==NULL)
	{
  	 	sprintf(m_message,"\nError:failed to allocate the LPPACKET structure."); 
	 	return;
	}	

	PacketInitPacket(m_sendLpPacket,(char*)m_sendBuffer, MAX_ETH_SIZE);

	/// Adapter to recieve messages 

	// WAN Side only
	PacketSetBuff(m_recvLpAdapter, RECV_BUF);
	
	//PacketSetReadTimeout(m_recvLpAdapter, 0);	// Blocking mode
	PacketSetReadTimeout(m_recvLpAdapter, 10000);	// 10 seconds
	
	//PacketSetHwFilter(m_recvLpAdapter,NDIS_PACKET_TYPE_DIRECTED);
	PacketSetHwFilter(m_recvLpAdapter,NDIS_PACKET_TYPE_PROMISCUOUS);

	if((m_recvLpPacket = PacketAllocatePacket())==NULL)
	{
		sprintf(m_message,"\nError:failed to allocate the LPPACKET structure."); 
		return;
	}
	PacketInitPacket(m_recvLpPacket,(char*)m_recvBuffer, sizeof(m_recvBuffer));
		
	strcpy(m_dhcpID, dhcpID);

	m_waitTime = 10; //seconds

	m_CDhcpFlag = true;

	memset(&m_dhcpGetInfo, 0, sizeof(struct dhcpGetInfo));

}




bool 
CDhcpcd::dhcpcdTempRecvThread()
{
	while(1)
	{
		if(PacketReceivePacket(m_recvLpAdapter,m_recvLpPacket,TRUE) == FALSE)
		{
			sprintf(m_message,"Error: PacketReceivePacket failed or Timeout");
			return false;
		}
		
		packetDelimiter(&m_dhcpGetInfo);
	}
}

unsigned int __stdcall _dhcpcdTempRecvThread(void* t) 
{
	CDhcpcd* dhcpcd = (CDhcpcd*)t;
	dhcpcd->dhcpcdTempRecvThread();

	_endthreadex(0);
	return 0;
}

bool
CDhcpcd::start(struct dhcpGetInfo *_dhcpGetInfo)
{
	///// start Temporary Recv Thread	

	m_dhcpcdTempRecvThreadHandle
		= (HANDLE)_beginthreadex(
			NULL, 0, &_dhcpcdTempRecvThread, this, 0, &m_dhcpcdTempRecvThreadID); 
	
	////////////////////////////////////////////////////////////////////	

	//char _packetMessage[342];

	///// make DISCOVER Message
	buildDiscoverMsg(m_sendBuffer);

	m_waitingPacketType = DHCP_OFFER;

	///// send DISCOVER Message
	// Release Preparation
	PacketInitPacket(m_sendLpPacket, m_sendBuffer, 342);
	// Release !
	BOOLEAN truefalse = PacketSendPacket(m_sendLpAdapter,m_sendLpPacket,TRUE);

#if 1

	SetEvent(m_dhcpcdLock_in);

	if (WaitForSingleObject(m_dhcpcdLock_out,INFINITE) != WAIT_OBJECT_0)// Wait INFINITE.
		return false;
#endif //1


	////////////////////////////////////////////////////////////////////

	///// Make REQUEST Message
	buildRequestMsg(m_sendBuffer, &m_dhcpGetInfo);

	m_waitingPacketType = DHCP_ACK;

	///// send REQUEST Message
	// Release Preparation
	PacketInitPacket(m_sendLpPacket, m_sendBuffer, 342);
	// Release !
	truefalse = PacketSendPacket(m_sendLpAdapter,m_sendLpPacket,TRUE);

	if (truefalse == false)
	{
		sprintf(m_message,"Error: cannot send REQUEST Message ");
		return false;
	}

#if 1
	SetEvent(m_dhcpcdLock_in);

	if (WaitForSingleObject(m_dhcpcdLock_out,INFINITE) != WAIT_OBJECT_0)// Wait INFINITE.
		return false;
#endif //1

	// DHCP_DECLINE 
	// revisit
	
	///// Set T1, T2 Timer
	__u32 _dummy;

	memcpy(&_dummy, m_dhcpGetInfo.leaseTime, 4);
	__u32 _leaseTime = ntohl(_dummy);
	
	m_dhcpGetInfo.dhcpLeaseTime = (__u32)_leaseTime; // second
	m_dhcpGetInfo.dhcpT1value = (__u32)(_leaseTime * 0.5);		// Seconds
	m_dhcpGetInfo.dhcpT2value = (__u32)(_leaseTime * 0.875);		// Seconds

	memcpy(_dhcpGetInfo, &m_dhcpGetInfo, sizeof(struct dhcpGetInfo));

	m_dhcpcdThreadHandle
		= (HANDLE)_beginthreadex(
			NULL, 0, &_dhcpcdThread, this, 0, &m_dhcpcdThreadID); 

	return true;
}


bool 
CDhcpcd::dhcpClientDemon()
{
	m_dhcpGetInfo.dhcpDemonStatus = true;

	while(1) // nest#0
	{
		bool nest0flag = true;
		int roopCounter = 0;

		Sleep(m_dhcpGetInfo.dhcpT2value * 1000);
		
		////////////////////////////////////////////////////////////////////
		///// Make REQUEST Message
		buildRequestMsg(m_sendBuffer, &m_dhcpGetInfo);

		int roopCount = 0;

		while(1) // nest#1
		{
			bool nest1flag = false;

			m_waitingPacketType = DHCP_ACK;
			// Packet Release !!

			// Release Preparation
			PacketInitPacket(m_sendLpPacket, m_sendBuffer, 342);
			// Release !
			BOOLEAN truefalse = PacketSendPacket(m_sendLpAdapter,m_sendLpPacket,TRUE);
			
			SetEvent(m_dhcpcdLock_in);

			if (WaitForSingleObject(m_dhcpcdLock_out,m_waitTime * 1000) != WAIT_OBJECT_0)// Wait 10 seconds
			{
				if (roopCount++ > 5) 
				{
					nest0flag = false;
					break;
				}
				Sleep((m_dhcpGetInfo.dhcpLeaseTime - m_dhcpGetInfo.dhcpT2value) / 5 * 1000);
			}

			// Check and set T1,T2 and other parameter again
			
			__u32 _dummy;
			
			memcpy(&_dummy, m_dhcpGetInfo.leaseTime, 4);
			__u32 _leaseTime = ntohl(_dummy);
			
			m_dhcpGetInfo.dhcpLeaseTime = (__u32)_leaseTime; // second
			m_dhcpGetInfo.dhcpT1value = (__u32)(_leaseTime * 0.5);		// Seconds
			m_dhcpGetInfo.dhcpT2value = (__u32)(_leaseTime * 0.875);		// Seconds
			
			break;
		}
		
		if (nest0flag == false)
			break;
	}
		
	m_dhcpGetInfo.dhcpDemonStatus = false;
	return false;
}

bool 
CDhcpcd::status(struct dhcpGetInfo *_dhcpGetInfo)
{
	memcpy(_dhcpGetInfo, &m_dhcpGetInfo, sizeof(struct dhcpGetInfo));

	return m_dhcpGetInfo.dhcpDemonStatus;
}

bool 
CDhcpcd::end(void)
{
	DWORD exitCode;
	GetExitCodeThread(m_dhcpcdThreadHandle, &exitCode);
	
	if (exitCode == STILL_ACTIVE)
	{
		TerminateThread(m_dhcpcdThreadHandle, exitCode);
		CloseHandle(m_dhcpcdThreadHandle);
	}

	return true;
}


CDhcpcd::~CDhcpcd()
{
	end();

	PacketFreePacket(m_sendLpPacket);
	PacketCloseAdapter(m_sendLpAdapter);

	PacketFreePacket(m_recvLpPacket);
	PacketCloseAdapter(m_recvLpAdapter);

	printf(".....CDhcpcd is destructed.\n");
}

